# -*- coding: utf-8 -*-
#
# FRETBursts - A single-molecule FRET burst analysis toolkit.
#
# Copyright (C) 2013-2016 The Regents of the University of California,
#               Antonino Ingargiola <tritemio@gmail.com>
#
"""
The module `burtlib_ext.py` (by default imported as `bext`) contains extensions
to `burstslib.py`. It can be though as a simple plugin system for FRETBursts.

Functions here defined operate on :class:`fretbursts.burstlib.Data()` objects
extending the functionality beyond the core functions and methods defined in
`burstlib`. This modularization allows to implement new functions without
overloading the :class:`fretbursts.burstlib.Data` with an high number
of non-core methods.

The type of functions here implemented are quite diverse. A short summary
follows.

* :func:`burst_search_and_gate` performs the AND-gate burst search taking
  intersection of the bursts detected in two photons streams.

* :func:`burst_data` returns a pandas DataFrame with burst data (one burst
  per row). Burst data includes sizes, duration, E, S, etc....

* :func:`bursts_fitter` and :func:`fit_bursts_kde_peak` help to build and
  fit histograms and KDEs for E or S.

* :func:`calc_mdelays_hist` computes the histogram of the m-delays
  distribution of photon intervals.

* :func:`moving_window_chunks`: slices the measurement using a moving-window
  (along the time axis). Used to follow or detect kinetics.

* :func:`join_data` joins different measurements to create a single
  "virtual" measurement from a series of measurements.

Finally a few functions deal with burst timestamps:

* :func:`get_burst_photons` returns a list of timestamps for each burst.
* :func:`ph_burst_stats` compute any statistics (for example mean or median)
  on the timestamps of each burst.
* :func:`asymmetry` returns a burst "asymmetry index" based on the difference
  between Donor and Acceptor timestamps.

"""
from __future__ import division, print_function, absolute_import
from builtins import range, zip

import numpy as np
from scipy.stats import erlang
from scipy.optimize import leastsq
import pandas as pd
import tables

from .ph_sel import Ph_sel
from . import background as bg
from .utils.misc import pprint, HistData, _is_list_of_arrays

from . import burstlib
from . import select_bursts
from . import fret_fit
from . import mfit

from .burstlib import isarray, Data
from .phtools.burstsearch import Bursts


def moving_window_startstop(start, stop, step, window=None):
    """Computes list of (start, stop) values defining a moving-window.

    Arguments:
        start, stop (scalars): range spanned by the moving window.
        step (scalar): window shift at each "step".
        window (scalar): window duration. If None, window = step.

    Returns:
        A list of (start, stop) values for the defined moving-window range.
    """
    if window is None:
        window = step
    # Reduce `stop` to an integer number of `step` and adjust for last window
    stop_corrected = (stop - stop % step) - window + step
    # differently from range, np.arange accepts also floats
    return [(t1, t1 + window) for t1 in np.arange(start, stop_corrected, step)]


def moving_window_dataframe(start, stop, step, window=None, time_zero=0):
    """Create a DataFrame for storing moving-window data.

    Create and return a DataFrame for storing columns of moving-window data.
    Three columns are initialize with "time axis" data: 'tstart', 'tstop'
    and 'tmean'. The returned DataFrame is typically used to store (in new
    columns) quantities as function of the moving time-window.
    Examples of such quantities are number of bursts, mean burst size/duration,
    fitted E peak position, etc.

    Arguments:
        start, stop (scalars): range spanned by the moving window.
        step (scalar): window shift at each "step".
        window (scalar): window duration. If None, window = step.

    Returns:
        DataFrame with 3 columns (tstart, tstop, tmean), one row for each
        window position.

    See also: :func:`moving_window_chunks`.
    """
    mw_slices = np.array(moving_window_startstop(start, stop, step, window))
    tstart = mw_slices[:, 0]
    tstop = mw_slices[:, 1]
    tmean = 0.5 * (tstart + tstop)
    df = pd.DataFrame(data=dict(tstart=tstart, tstop=tstop, tmean=tmean))
    return df - time_zero


def moving_window_chunks(dx, start, stop, step, window=None,
                         time_zero=0):
    """Return a list of Data object, each containing bursts in one time-window.

    Each returned Data object contains only bursts lying in the current
    time-window. Additionally, the start/stop values of current time-window
    are saved in `Data`'s attributes: name, slice_tstart and slice_tstop.

    Arguments:
        dx (Data): the Data() object to be sliced with a moving window.
        start, stop (scalars): time-range in seconds spanned by the
            moving window.
        step (scalar): window time-shift at each step.
        window (scalar): window duration. If None, window = step.
        time_zero (scalar): shift the start/stop times saved in the Data
            attributes so that "time zero" falls at `time_zero` seconds.
            Default 0, no shift.

    Returns:
        A list of Data objects, one for each window position.

    See also: :func:`moving_window_dataframe`.
    """
    time_slices = moving_window_startstop(start, stop, step, window)
    dx_slices = []
    for t1, t2 in time_slices:
        dx_slice = dx.select_bursts(select_bursts.time,
                                    time_s1=t1, time_s2=t2)
        dx_slice.name = 'Slice %d:%d s' % (t1 - time_zero, t2 - time_zero)
        dx_slice.add(slice_tstart=t1 - time_zero, slice_tstop=t2 - time_zero)
        dx_slices.append(dx_slice)
    return dx_slices


def calc_mean_lifetime(dx, t1=0, t2=np.inf, ph_sel=Ph_sel('all')):
    """Compute the mean lifetime in each burst.

    Arguments:
        t1, t2 (floats): min and max value (in TCSPC bin units) for the
            nanotime to be included in the mean
        ph_sel (Ph_sel object): object defining the photon selection.
            See :mod:`fretbursts.ph_sel` for details.

    Returns:
        List of arrays of per-burst mean lifetime. One array per channel.
    """
    mean_lifetimes = []

    for bursts, nanot, mask in zip(dx.mburst, dx.nanotimes,
                                   dx.iter_ph_masks(ph_sel)):
        selection = (nanot > t1) * (nanot < t2)
        # Select photons in ph_sel AND with nanotime in [t1, t2]
        if isarray(mask):
            selection *= mask
        mean_lifetimes.append(
            burstlib.burst_ph_stats(nanot, bursts, mask=selection,
                                    func=np.mean) - t1)

    return mean_lifetimes


def _store_bg_data(store, base_name, min_ph_delays_us, best_bg, best_th,
                   BG_data, BG_data_e):
    if not base_name.endswith('/'):
        base_name = base_name + '/'
    store_name = store.filename
    group_name = '/' + base_name[:-1]
    store.create_carray(group_name, 'min_ph_delays_us', obj=min_ph_delays_us,
                        createparents=True)
    for ph_sel, values in BG_data.items():
        store.create_carray(group_name, str(ph_sel), obj=values)
    for ph_sel, values in BG_data_e.items():
        store.create_carray(group_name, str(ph_sel) + '_err', obj=values)
    store.close()
    store = pd.HDFStore(store_name)
    store[base_name + 'best_bg'] = best_bg
    store[base_name + 'best_th'] = best_th
    store.close()


def _load_bg_data(store, base_name, ph_streams):
    if not base_name.endswith('/'):
        base_name = base_name + '/'
    store_name = store.filename
    group_name = '/' + base_name[:-1]
    min_ph_delays = store.get_node(group_name, 'min_ph_delays_us')[:]
    BG_data = {}
    for ph_sel in ph_streams:
        BG_data[ph_sel] = store.get_node(group_name, str(ph_sel))[:]
    BG_data_e = {}
    for ph_sel in ph_streams:
        BG_data_e[ph_sel] = store.get_node(group_name, str(ph_sel) + '_err')[:]
    store.close()
    store = pd.HDFStore(store_name)
    best_bg = store[base_name + 'best_bg']
    best_th = store[base_name + 'best_th']
    store.close()
    return best_th, best_bg, BG_data, BG_data_e, min_ph_delays


def calc_bg_brute_cache(dx, min_ph_delay_list=None, return_all=False,
                        error_metrics='KS', force_recompute=False):
    """Compute background for all the ch, ph_sel and periods caching results.

    This function performs a brute-force search of the min ph delay
    threshold. The best threshold is the one the minimizes the error
    function. The best background fit is the rate fitted using the
    best threshold.

    Results are cached to disk and loaded transparently when needed.
    The cache file is an HDF5 file named `dx.fname[:-5] + '_BKG.hdf5'`.

    Arguments:
        min_ph_delay_list (sequence): sequence of values used for the
            brute-force search. Background and error will be computed
            for each value in `min_ph_delay_list`.
        return_all (bool): if True return all the fitted backgrounds and
            error functions. Default False.
        error_metrics (string): Specifies the error metric to use.
            See :func:`fretbursts.background.exp_fit` for more details.
        force_recompute (bool): if True, recompute results even if a cache
            is found.

    Returns:
        Two arrays with best threshold (us) and best background. If
        `return_all = True` also returns the dictionaries containing all the
        fitted backgrounds and errors.
    """
    if min_ph_delay_list is None:
        min_ph_delay_list = np.arange(100, 8500, 100)
    else:
        min_ph_delay_list = np.asfarray(min_ph_delay_list)

    base_name = 'bg_%s_%ds_%s/' % (dx.bg_fun_name, dx.bg_time_s, error_metrics)
    store_fname = dx.fname[:-5] + '_BKG.hdf5'
    comp_filter = tables.Filters(complevel=6, complib='zlib')
    store = tables.open_file(store_fname, mode='a', filters=comp_filter)
    loaded = False
    if '/' + base_name + 'min_ph_delays_us' in store:
        Th = store.get_node('/', base_name + 'min_ph_delays_us').read()
        if np.all(Th == min_ph_delay_list) and not force_recompute:
            print(' - Loading BG from cache')
            res = _load_bg_data(store, base_name, dx.ph_streams)
            loaded = True
    if not loaded:
        print(' - Computing BG')
        res = calc_bg_brute(dx, min_ph_delay_list=min_ph_delay_list,
                            return_all=True, error_metrics=error_metrics)
        best_th, best_bg, BG_data, BG_data_e, min_ph_delays = res
        _store_bg_data(store, base_name, min_ph_delays, best_bg, best_th,
                       BG_data, BG_data_e)

    if return_all:
        return res
    else:
        return res[:-3]


def calc_bg_brute(dx, min_ph_delay_list=None, return_all=False,
                  error_metrics='KS'):
    """Compute background for all the ch, ph_sel and periods.

    This function performs a brute-force search of the min ph delay
    threshold. The best threshold is the one the minimizes the error
    function. The best background fit is the rate fitted using the
    best threshold.

    Arguments:
        min_ph_delay_list (sequence): sequence of values used for the
            brute-force search. Background and error will be computed
            for each value in `min_ph_delay_list`.
        return_all (bool): if True return all the fitted backgrounds and
            error functions. Default False.
        error_metrics (string): Specifies the error metric to use.
            See :func:`fretbursts.background.exp_fit` for more details.

    Returns:
        Two arrays with best threshold (us) and best background. If
        `return_all = True` also returns the dictionaries containing all the
        fitted backgrounds and errors.
    """
    if min_ph_delay_list is None:
        min_ph_delay_list = np.arange(100, 8500, 100)
    else:
        min_ph_delay_list = np.asfarray(min_ph_delay_list)

    ph_sel_labels = [str(p) for p in dx.ph_streams]

    BG_data, BG_data_e = {}, {}

    index = pd.MultiIndex.from_product([list(range(dx.nch)), ph_sel_labels],
                                       names=['CH', 'ph_sel'])
    best_th = pd.DataFrame(index=index, columns=np.arange(dx.nperiods))
    best_th.columns.name = 'period'
    best_th.sortlevel(inplace=True)
    best_bg = best_th.copy()

    for ph_sel in dx.ph_streams:
        # Compute BG and error for all ch, periods and thresholds
        # Shape: (nch, nperiods, len(thresholds))
        BG_data[ph_sel], BG_data_e[ph_sel] = bg.fit_varying_min_delta_ph(
            dx, min_ph_delay_list, bg_fit_fun=bg.exp_fit, ph_sel=ph_sel,
            error_metrics=error_metrics)

        # Compute the best Th and BG estimate for all ch and periods
        for ich in range(dx.nch):
            for period in range(dx.nperiods):
                b = BG_data_e[ph_sel][ich, period, :]
                i_best_th = b[-np.isnan(b)].argmin()
                best_th.loc[(ich, str(ph_sel)), period] = \
                    min_ph_delay_list[i_best_th]
                best_bg.loc[(ich, str(ph_sel)), period] = \
                    BG_data[ph_sel][ich, period, i_best_th]
    if return_all:
        return best_th, best_bg, BG_data, BG_data_e, min_ph_delay_list
    else:
        return best_th, best_bg


def burst_data(dx, include_bg=False, include_ph_index=False,
               skip_ch=None):
    """Return a table (`pd.DataFrame`) of burst data (one row per burst).

    Columns include:

    - *E* and *S*: FRET and stoichiometry for each burst.
    - *nd*, *na*, *naa*, *nda*: burst counts in DexDem, DexAem, AexAem and
      AexDem photon streams.
    - *t_start*, *t_stop*: time (in seconds) of first and last photon inside
      the burst
    - *width_ms*: burst duration in milliseconds
    - *size_raw*: the total uncorrected burst counts in the photon stream
      used for burst search

    Optional columns include:

    - *i_start*, *i_stop*: index of burst start and stop relative to the
      original timestamps array (requires `include_ph_index=True`)
    - *bg_dd*, *bg_ad*, *bg_aa*, *bg_da*: background contribution in the
      DexDem, DexAem, AexAem, AexDem photon streams (requires
      `include_bg=True`).

    If the peak photon-counts in each bursts has been computed (see
    :meth:`fretbursts.burstlib.Data.calc_max_rate`), it will
    be included as a column called *max_rate*.

    Arguments:
        include_bg (bool): if True includes additional columns for burst
            background (see above). Default False.
        include_ph_index (bool): if True includes additional two columns for
            index of first and last timestamp in each burst. Default False.
        skip_ch (list or None): List of channels to skip if measurement is
            multispot.

    Return:
        A pandas's DataFrame containing burst data (one row per burst).
    """
    def _burst_data_ich(dx, ich, include_bg=False, include_ph_index=False):
        bursts_ich = dx.burst_data_ich(ich=ich)
        if not include_ph_index:
            bursts_ich.pop('i_start')
            bursts_ich.pop('i_stop')
        if not include_bg:
            bursts_ich = {k: v for k, v in bursts_ich.items()
                          if not k.startswith('bg')}
        return pd.DataFrame.from_dict(bursts_ich)

    if skip_ch is None:
        skip_ch = []
    kws = dict(include_bg=include_bg, include_ph_index=include_ph_index)
    bursts = {}
    for ich in range(dx.nch):
        if ich not in skip_ch and dx.num_bursts[ich] > 0:
            bursts[ich] = _burst_data_ich(dx, ich=ich, **kws)
            bursts[ich]['spot'] = ich
    if dx.nch > 1:
        bursts = pd.concat(bursts, ignore_index=True)
    else:
        bursts = bursts[0]
    return bursts


def fit_bursts_kde_peak(dx, burst_data='E', bandwidth=0.03, weights=None,
                        gamma=1, add_naa=False, x_range=(-0.1, 1.1),
                        x_ax=None, save_fitter=True):
    """Fit burst data (typ. E or S) by finding the KDE max on all the channels.

    Arguments:
        dx (Data): `Data` object containing the FRET data
        burst_data (string): name of burst-data attribute (i.e 'E' or 'S').
        bandwidth (float): bandwidth for the Kernel Density Estimation
        weights (string or None): kind of burst-size weights.
            See :func:`fretbursts.fret_fit.get_weights`.
        gamma (float): gamma factor passed to `get_weights()`.
        add_naa (bool): if True adds `naa` to the burst size.
        save_fitter (bool): if True save the `MultiFitter` object in the
            `dx` object with name: burst_data + '_fitter'.
        x_range (tuple of floats): min-max range where to search for the peak.
            Used to select a single peak in a multi-peaks distribution.
        x_ax (array or None): x-axis used to evaluate the Kernel Density

    Returns:
        An array of max peak positions (one per ch). If the number of
        channels is 1 returns a scalar.
    """
    if x_ax is None:
        x_ax = np.arange(-0.2, 1.2, 0.0002)

    fitter = bursts_fitter(dx, burst_data=burst_data, save_fitter=save_fitter,
                           weights=weights, gamma=gamma, add_naa=add_naa)
    fitter.calc_kde(bandwidth=bandwidth)
    fitter.find_kde_max(x_ax, xmin=x_range[0], xmax=x_range[1])
    KDE_max_mch = fitter.kde_max_pos
    if dx.nch == 1:
        KDE_max_mch = KDE_max_mch[0]
    return KDE_max_mch

def bursts_fitter(dx, burst_data='E', save_fitter=True,
                  weights=None, gamma=1, add_naa=False, skip_ch=None,
                  binwidth=None, bandwidth=None, model=None, verbose=False):
    """Create a mfit.MultiFitter object (for E or S) add it to `dx`.

    A MultiFitter object allows to fit multi-channel data with the same
    model.

    Arguments:
        dx (Data): `Data` object containing the FRET data
        save_fitter (bool): if True save the `MultiFitter` object in the
            `dx` object with name: burst_data + '_fitter'.
        burst_data (string): name of burst-data attribute (i.e 'E' or 'S').
        weights (string or None): kind of burst-size weights.
            See :func:`fretbursts.fret_fit.get_weights`.
        gamma (float): gamma factor passed to `get_weights()`.
        add_naa (bool): if True adds `naa` to the burst size.
        binwidth (float or None): bin width used to compute the histogram.
            If None the histogram is not computed.
        bandwidth (float or None): bandwidth used to compute the KDE
            If None the KDE is not computed.
        model (lmfit.Model object or None): lmfit Model used for histogram
            fitting. If None the histogram is not fitted.
        verbose (bool): if False avoids printing any output.

    Returns:
        The `mfit.MultiFitter` object with the specified burst-size weights.
    """
    assert burst_data in dx
    fitter = mfit.MultiFitter(dx[burst_data], skip_ch=skip_ch)
    if weights is None or _is_list_of_arrays(weights):
        # If no weights or precomputed weights
        fitter.weights = weights
    else:
        # Otherwise build the kwargs to compute weights through get_weights
        weight_kwargs = dict(weights=weights, gamma=gamma,
                             nd=dx.nd, na=dx.na)
        if add_naa:
            weight_kwargs['naa'] = dx.naa
        if weights == 'brightness':
            weight_kwargs['widths'] = [mb.width*dx.clk_p for mb in dx.mburst]
        fitter.set_weights_func(weight_func=fret_fit.get_weights,
                                weight_kwargs=weight_kwargs)
    if bandwidth is not None:
        fitter.calc_kde(bandwidth)
    if binwidth is not None:
        fitter.histogram(binwidth=binwidth)
        if model is not None:
            fitter.fit_histogram(model=model, verbose=verbose)
    if save_fitter:
        # Save fitter variables as normal attributes (not with Data.add)
        # so they are not copied when copying Data (e.g. burst selection)
        setattr(dx, burst_data + '_fitter', fitter)
        setattr(dx, burst_data + '_weights', (weights, float(gamma), add_naa))
    return fitter

def _get_bg_distrib_erlang(d, ich=0, m=10, ph_sel=Ph_sel('all'),
                           period=(0, -1)):
    """Return a frozen (scipy) erlang distrib. with rate equal to the bg rate.
    """
    assert ph_sel in [Ph_sel('all'), Ph_sel(Dex='Dem'), Ph_sel(Dex='Aem')]

    # Compute the BG distribution
    if ph_sel == Ph_sel('all'):
        bg_ph = d.bg_dd[ich] + d.bg_ad[ich]
    elif ph_sel == Ph_sel(Dex='Dem'):
        bg_ph = d.bg_dd[ich]
    elif ph_sel == Ph_sel(Dex='Aem'):
        bg_ph = d.bg_ad[ich]

    rate_ch_kcps = bg_ph[period[0]:period[1]+1].mean()/1e3  # bg rate in kcps
    bg_dist = erlang(a=m, scale=1./rate_ch_kcps)
    return bg_dist

def _get_bg_erlang(d, ich=0, m=10, ph_sel=Ph_sel('all'), period=0):
    """Return a frozen (scipy) erlang distrib. with rate equal to the bg rate.
    """
    bg_rate = d.bg_from(ph_sel=ph_sel)[ich][period]
    #bg_rate_kcps = bg_rate*1e-3
    bg_dist = erlang(a=m, scale=1./bg_rate)
    return bg_dist

def histogram_mdelays(d, ich=0, m=10, ph_sel=Ph_sel('all'),
                      binwidth=1e-3, dt_max=10e-3, bins=None,
                      inbursts=False):
    """Compute histogram of m-photons delays (or waiting times).

    Arguments:
        dx (Data object): contains the burst data to process.
        ich (int): the channel number. Default 0.
        m (int): number of photons used to compute each delay.
        ph_sel (Ph_sel object): photon selection to use.
        inbursts (bool): if True, compute the histogram with only
            photons in bursts.

    Returns:
        A `HistData` object containing the computed histogram.
    """
    if bins is None:
        bins = np.arange(0, dt_max, binwidth)

    if not inbursts:
        ph = d.get_ph_times(ich=ich, ph_sel=ph_sel)
    else:
        ph = d.ph_in_bursts_ich(ich=ich, ph_sel=ph_sel)

    ph_mdelays = np.diff(ph[::m])*d.clk_p
    hist = HistData(*np.histogram(ph_mdelays, bins=bins))
    return hist

def calc_mdelays_hist(d, ich=0, m=10, period=(0, -1), bins_s=(0, 10, 0.02),
                      ph_sel=Ph_sel('all'), bursts=False, bg_fit=True,
                      bg_F=0.8):
    """Compute histogram of m-photons delays (or waiting times).

    Arguments:
        dx (Data object): contains the burst data to process.
        ich (int): the channel number. Default 0.
        m (int): number of photons used to compute each delay.
        period (int or 2-element tuple): index of the period to use. If
            tuple, the period range between period[0] and period[1]
            (included) is used.
        bins_s (3-element tuple): start, stop and step for the bins
        ph_sel (Ph_sel object): photon selection to use.

    Returns:
        Tuple of values:
        * bin_x (array): array of bins centers
        * histograms_y (array): arrays of histograms, contains 1 or 2
          histograms (when `bursts` is False or True)
        * bg_dist (random distribution): erlang distribution with same
          rate as background (kcps)
        * a, rate_kcps (floats, optional): amplitude and rate for an
          Erlang distribution fitted to the histogram for
          bin_x > bg_mean*bg_F. Returned only if `bg_fit` is True.
    """
    assert ph_sel in [Ph_sel('all'), Ph_sel(Dex='Dem'), Ph_sel(Dex='Aem')]
    if np.size(period) == 1: period = (period, period)
    periods = slice(d.Lim[ich][period[0]][0], d.Lim[ich][period[1]][1] + 1)
    bins = np.arange(*bins_s)

    if ph_sel == Ph_sel('all'):
        ph = d.ph_times_m[ich][periods]
        if bursts:
            phb = ph[d.ph_in_burst[ich][periods]]
    elif ph_sel == Ph_sel(Dex='Dem'):
        donor_ph_period = -d.A_em[ich][periods]
        ph = d.ph_times_m[ich][periods][donor_ph_period]
        if bursts:
            phb = ph[d.ph_in_burst[ich][periods][donor_ph_period]]
    elif ph_sel == Ph_sel(Dex='Aem'):
        accept_ph_period = d.A_em[ich][periods]
        ph = d.ph_times_m[ich][periods][accept_ph_period]
        if bursts:
            phb = ph[d.ph_in_burst[ich][periods][accept_ph_period]]

    ph_mdelays = np.diff(ph[::m])*d.clk_p*1e3        # millisec
    if bursts:
        phb_mdelays = np.diff(phb[::m])*d.clk_p*1e3  # millisec
        phb_mdelays = phb_mdelays[phb_mdelays < 5]

    # Compute the PDF through histograming
    hist_kwargs = dict(bins=bins, normed=True)
    mdelays_hist_y, _ = np.histogram(ph_mdelays, **hist_kwargs)
    bin_x = bins[:-1] + 0.5*(bins[1] - bins[0])
    if bursts:
        mdelays_b_hist_y, _ = np.histogram(phb_mdelays, **hist_kwargs)
        mdelays_b_hist_y *= phb_mdelays.size/ph_mdelays.size
    if bursts:
        histograms_y = np.vstack([mdelays_hist_y, mdelays_b_hist_y])
    else:
        histograms_y = mdelays_hist_y

    results = [bin_x, histograms_y]

    # Compute the BG distribution
    bg_dist = _get_bg_distrib_erlang(d, ich=ich, m=m, period=period,
                                     ph_sel=ph_sel)
    bg_mean = bg_dist.mean()
    results.append(bg_dist)

    if bg_fit:
        ## Fitting the BG portion of the PDF to an Erlang
        _x = bin_x[bin_x > bg_mean*bg_F]
        _y = mdelays_hist_y[bin_x > bg_mean*bg_F]
        fit_func = lambda x, a, rate_kcps: a*erlang.pdf(x, a=m,
                                                        scale=1./rate_kcps)
        err_func = lambda p, x, y: fit_func(x, p[0], p[1]) - y
        p, flag = leastsq(err_func, x0=[0.9, 3.], args=(_x, _y))
        print(p, flag)
        a, rate_kcps = p
        results.extend([a, rate_kcps])

    return results

def burst_data_period_mean(dx, burst_data):
    """Compute mean `burst_data` in each period.

    Arguments:
        dx (Data object): contains the burst data to process
        burst_data (list of arrays): one array per channel, each array
            has one element of "burst data" per burst.

    Returns:
        2D of arrays with shape (nch, nperiods).

    Example:
        burst_period_mean(dx, dx.nt)
    """
    mean_burst_data = np.zeros((dx.nch, dx.nperiods))
    for ich, (b_data_ch, period) in enumerate(zip(burst_data, dx.bp)):
        for iperiod in range(dx.nperiods):
            mean_burst_data[ich, iperiod] = b_data_ch[period == iperiod].mean()
    return mean_burst_data

def join_data(d_list, gap=0):
    """Joins burst data of different measurements in a single `Data` object.

    Merge a list of `Data` objects (i.e. a set of different measurements)
    into a single `Data` object containing all the bursts (like it was a
    single acquisition).
    The `Data` objects to be merged need to already contain burst data.
    The input `Data` objects are required to have undergone background
    estimation (all with the same background period) and burst search.
    For each measurement, the time of burst start is offset by the duration
    of the previous measurement + an additional `gap` (which is 0 by
    default).

    The index of the first/last photon in the burst (`istart` and `iend`)
    are kept unmodified and refer to the original timestamp array.
    The timestamp arrays are not copied: the new `Data` object will
    not contain any timestamp arrays (ph_times_m). This may cause errors when
    calling functions that require the timestamps data such as burst search.

    The background arrays (bg, bg_dd, etc...) are concatenated. The burst
    attribute `bp` is updated to refer to these new concatenated arrays.
    The attributes `Lim` and `Ph_p` are concatenated and left unchanged.
    Therefore different sections will refer to different original timestamp
    arrays. The retuned `Data` object will have a new attribute `i_origin`,
    containing, for each burst, the index of the original data object
    in the list.

    Arguments:
        d_list (list of Data objects): the list of measurements to concatenate.
        gap (float): the time delay (or gap) in seconds to add to each
            concatenated measurement.

    Returns:
        A `Data` object containing bursts from the all the objects in `d_list`.
        This object will not contain timestamps, therefore it is possible
        to perform burst selections but not a new burst serach.

    Example:
        If `d1` and `d2` are two measurements to concatenate::

            file_list = ['filename1', 'filename2']
            d_list = [loader.photon_hdf5(f) for f in file_list]

            for dx in d_list:
                loader.alex_apply_period(dx)
                dx.calc_bg(bg.exp_fit, time_s=30, tail_min_us='auto', F_bg=1.7)
                dx.burst_search()

            d_merged = bext.join_data(d_list)

        `d_merged` will contain bursts from both input files.

    """
    from itertools import islice

    nch = d_list[0].nch
    bg_time_s = d_list[0].bg_time_s
    for d in d_list:
        assert d.nch == nch
        assert d.bg_time_s == bg_time_s

    new_d = Data(**d_list[0])
    new_d.delete('ph_times_m')

    # Set the bursts fields by concatenation along axis = 0
    for name in Data.burst_fields:
        if name in new_d:
            empty = Bursts.empty() if name is 'mburst' else np.array([])
            new_d.add(**{name: [empty]*nch})
            concatenate = Bursts.merge if name == 'mburst' else np.concatenate

            for ich in range(nch):
                new_size = np.sum((d.mburst[ich].num_bursts for d in d_list))
                if new_size == 0:
                    continue  # -> No bursts in this ch

                value = concatenate([d[name][ich] for d in d_list])
                new_d[name][ich] = value
                assert new_d[name][ich].size == new_size

    # Set the background fields by concatenation along axis = 0
    new_nperiods = np.sum((d.nperiods for d in d_list))
    for name in ('Lim', 'Ph_p'):
        if name in new_d:
            new_d.add(**{name: []})
            for ich in range(nch):
                value = np.concatenate([d[name][ich] for d in d_list])
                new_d[name].append(value)
                assert new_d[name][ich].shape[0] == new_nperiods
    if 'bg' in new_d:
        new_d.add(bg={})
        for sel in d.bg:
            new_d.bg[sel] = []
            for ich in range(nch):
                value = np.concatenate([d.bg[sel][ich] for d in d_list])
                new_d.bg[sel].append(value)
                assert new_d.bg[sel][ich].shape[0] == new_nperiods

    # Set the i_origin burst attribute
    new_d.add(i_origin=[])
    for ich in range(nch):
        i_origin_ch = np.concatenate([i_d * np.ones(d.num_bursts[ich])
                                      for i_d, d in enumerate(d_list)])
        new_d.i_origin.append(i_origin_ch)

    # Update the `bp` attribute to refer to the background period in
    # the new concatenated background arrays.
    sum_nperiods = np.cumsum([d.nperiods for d in d_list])
    for i_d, d in islice(enumerate(d_list), 1, None):
        for ich in range(nch):
            # Burst "slice" in new_d coming from current d
            b_mask = new_d.i_origin[ich] == i_d
            # Add the nperiods of all the previous measurements
            new_d.bp[ich][b_mask] = new_d.bp[ich][b_mask] + sum_nperiods[i_d-1]

    # Modify the new mburst so the time of burst start/end is monotonic
    offset_clk = 0
    iburst_start = d_list[0].num_bursts.copy()
    for d_i in d_list[1:]:
        offset_clk += int((d_i.time_max + gap) / d_i.clk_p)
        for ich in range(nch):
            if new_d.mburst[ich].num_bursts == 0:
                continue
            mask = slice(iburst_start[ich],
                         iburst_start[ich] + d_i.num_bursts[ich])
            bursts_ch_i = new_d.mburst[ich][mask]
            bursts_ch_i.start += offset_clk
            bursts_ch_i.stop += offset_clk
        iburst_start += d_i.num_bursts

    return new_d


def burst_search_and_gate(dx, F=6, m=10, min_rate_cps=None, c=-1,
                          ph_sel1=Ph_sel(Dex='DAem'),
                          ph_sel2=Ph_sel(Aex='Aem'), compact=False, mute=False):
    """Return a Data object containing bursts obtained by and-gate burst-search.

    The and-gate burst search is a composition of 2 burst searches performed
    on different photon selections. The bursts in the and-gate burst search
    are the overlapping bursts in the 2 initial burst searches, and their
    duration is the intersection of the two overlapping bursts.

    By default the 2 photon selections are D+A photons during D excitation
    (`Ph_sel(Dex='DAem')`) and A photons during A excitation
    (`Ph_sel(Aex='Aex')`).

    Arguments:
        dx (Data object): contains the data on which to perform the burst
            search. Background estimation must be performed before the search.
        F (float or tuple): burst search parameter F. If it is a 2-element
            tuple, specifies F separately for `ph_sel1` and `ph_sel2`.
        m (int or tuple): Burst search parameter m. If it is a 2-element
            tuple, specifies m separately for `ph_sel1` and `ph_sel2`.
        min_rate_cps (float or tuple): min. rate in cps for burst detection.
            If not None, min_rate_cps overrides any value passed in `F`.
            If a 2-element tuple specifies min_rate_cps separately for
            `ph_sel1` and `ph_sel2`. In multispot data, it can also be an
            array (or a 2-tuple or arrays) with size equal to the number of
            spots.
        c (float): parameter used set the definition of the rate estimatator.
            See `c` parameter in :meth:`.burstlib.Data.burst_search`
            for details.
        ph_sel1 (Ph_sel object): photon selections used for bursts search 1.
        ph_sel2 (Ph_sel object): photon selections used for bursts search 2.
        mute (bool): if True nothing is printed. Default: False.

    Return:
        A new `Data` object containing bursts from the and-gate search.

    See also :meth:`fretbursts.burstlib.Data.burst_search`.
    """
    dx_d = dx.copy(mute=mute)
    dx_a = dx.copy(mute=mute)
    dx_and = dx.copy(mute=mute)
    dx_and.delete_burst_data()

    def _get_args(x):
        try:
            assert len(x) == 2
            x1, x2 = x
        except TypeError:
            x1 = x2 = x
        return x1, x2

    m1, m2 = _get_args(m)
    F1, F2 = _get_args(F)
    min_rate_cps1, min_rate_cps2 = _get_args(min_rate_cps)

    dx_d.burst_search(L=m1, m=m1, F=F1, min_rate_cps=min_rate_cps1, c=c,
                      ph_sel=ph_sel1, compact=compact, computefret=False,
                      mute=mute)
    dx_a.burst_search(L=m2, m=m2, F=F2, min_rate_cps=min_rate_cps2, c=c,
                      ph_sel=ph_sel2, compact=compact, computefret=False,
                      mute=mute)

    mburst_and = []
    for bursts_d, bursts_a in zip(dx_d.mburst, dx_a.mburst):
        mburst_and.append(bursts_d.and_gate(bursts_a))

    dx_and.add(mburst=mburst_and)

    pprint(" - Calculating burst periods ...", mute)
    dx_and._calc_burst_period()                       # writes bp
    pprint("[DONE]\n", mute)

    # Note: dx_and.bg_bs will not be meaningful
    dx_and.add(m=m, L=m, F=F, P=None, min_rate_cps=min_rate_cps,
               ph_sel='AND-gate')
    dx_and.add(bg_corrected=False, leakage_corrected=False,
               dir_ex_corrected=False, dithering=False)

    pprint(" - Counting D and A ph and calculating FRET ... \n", mute)
    dx_and.calc_fret(count_ph=True, corrections=True, mute=mute)
    pprint("   [DONE Counting D/A]\n", mute)

    return dx_and


##
#  Burst asymmetry
#

def get_burst_photons(d, ich=0, ph_sel=Ph_sel('all')):
    """Return a list of arrays of photon timestamps in each burst.

    Arguments:
        d (Data): Data() object
        ich (int): channel index
        ph_sel (Ph_sel): photon selection. It allows to select timestamps
            from a specific photon selection. Example ph_sel=Ph_sel(Dex='Dem').
            See :mod:`fretbursts.ph_sel` for details.

    Returns:
        A list of arrays of photon timestamps (one array per burst).
    """
    bursts = d.mburst[ich]
    i_start, i_end = bursts.istart, bursts.istop

    ph_times = d.get_ph_times(ich)
    burst_slices = [slice(i1, i2 + 1) for i1, i2 in zip(i_start, i_end)]
    burst_photons = [ph_times[slice_i] for slice_i in burst_slices]

    if ph_sel != Ph_sel('all'):
        ph_times_mask = d.get_ph_mask(ich, ph_sel=ph_sel)
        photon_masks = [ph_times_mask[slice_i] for slice_i in burst_slices]
        burst_photons = [ph[mask] for ph, mask in zip(burst_photons,
                                                      photon_masks)]
    return burst_photons

def ph_burst_stats(d, ich=0, func=np.mean, ph_sel=Ph_sel('all')):
    """Applies function `func` to the timestamps of each burst.

    Arguments:
        d (Data): Data() object
        ich (int): channel index
        func (function): a function that take an array of burst-timestamps
            and return a scalar. Default `numpy.mean`.
        ph_sel (Ph_sel): photon selection. It allows to select timestamps
            from a specific photon selection. Default Ph_sel('all').
            See :mod:`fretbursts.ph_sel` for details.

    Returns:
        An array containing per-burst timestamp statistics.
    """
    burst_photons = get_burst_photons(d, ich, ph_sel=ph_sel)
    stats = [func(times) for times in burst_photons]
    return np.array(stats)

def asymmetry(dx, ich=0, func=np.mean, dropnan=True):
    """Compute an asymmetry index for each burst in channel `ich`.

    It computes each burst the difference func({t_D}) - func({t_A})
    where `func` is a function (default `mean`) that computes some statistics
    on the timestamp and {t_D} and {t_A} are the sets of D or A timestamps
    in a bursts (during D excitation).

    Arguments:
        d (Data): Data() object
        ich (int): channel index
        func (function): the function to be used to extract D and A photon
            statistics in each bursts.

    Returns:
        An arrays of photon timestamps (one array per burst).
    """
    stats_d = ph_burst_stats(dx, ich=ich, func=func, ph_sel=Ph_sel(Dex='Dem'))
    stats_a = ph_burst_stats(dx, ich=ich, func=func, ph_sel=Ph_sel(Dex='Aem'))

    #b_size = d.burst_sizes(ich, add_naa=False)
    #b_width = d.mburst[ich].width
    burst_asym = (stats_d - stats_a)*dx.clk_p*1e3
    if dropnan:
        burst_asym = burst_asym[-np.isnan(burst_asym)]
    return burst_asym



#
# FRETBursts - A single-molecule FRET burst analysis toolkit.
#
# Copyright (C) 2014 Antonino Ingargiola <tritemio@gmail.com>
#
"""
Functions to check the version of a repository by querying git.
"""

from __future__ import print_function
import os
from distutils.spawn import find_executable
from subprocess import check_output, call


# GIT_PATHS is a list of paths for the git executables
# In the following we iterate through the list until we find a valid executable.
GIT_PATHS = []

# Set here the git executable path (if it is not the system path)
#GIT_PATHS.append(r'C:\git.exe')

# Add system path and common locations to the pool of git paths
system_git = find_executable('git')  # on windows it adds '.exe'
if system_git is not None:
    GIT_PATHS.append(system_git)
GIT_PATHS.append((os.environ.get('homepath', '') + \
        r'\AppData\Local\Atlassian\SourceTree\git_local\bin\git.exe'))

def find_git():
    git_path = None
    for git_path_i in GIT_PATHS:
        if git_path_i is not None:
            if os.path.isfile(git_path_i) and os.access(git_path_i, os.X_OK):
                git_path = git_path_i
                break
    return git_path

# GIT_PATH is the default git executable path used by functions in this module
# unless the `git_path` argument is passed.
GIT_PATH = find_git()


def find_sources_dir():
    """
    Returns FRETBursts sources folder from `HOME/.fretbursts` or `./`
    """
    if os.name == 'posix':
        # Linux or Mac
        HOME = os.environ['HOME'] + '/'
    elif os.name == 'nt':
        # Windows
        HOME = os.environ['HOMEPATH'] + '/'
    else:
        raise OSError ("Operating system not recognized (%s)." % os.name)

    config_file_name = '.fretbursts'
    FRETBURSTS_DIR = './'
    with open(HOME + config_file_name) as f:
        FRETBURSTS_DIR = f.read().strip()

    return FRETBURSTS_DIR

def git_path_valid(git_path=None):
    """
    Check whether the git executable is found.
    """
    if git_path is None and GIT_PATH is None:
        return False
    if git_path is None: git_path = GIT_PATH
    try:
        call([git_path, '--version'])
        return True
    except OSError:
        return False

def call_git(git_path, *git_args):
    """Call git with passed arguments from within FRETBursts source folder.

    Returns the output printed by the git call.
    """
    if git_path is None:
        git_path = GIT_PATH
    fretbursts_dir = find_sources_dir()
    curr_dir = os.path.abspath('./')
    os.chdir(fretbursts_dir)
    output = check_output([git_path] + list(git_args))
    os.chdir(curr_dir)
    return output

def get_git_version(git_path=None):
    """
    Get the Git version.
    """
    git_version = call_git(git_path, "--version").split()[2]
    return git_version

def get_status(git_path=None):
    """
    Returns a string listing all the uncommitted changes in the working dir.
    """
    output = call_git(git_path, "status", "--porcelain")
    return output

def check_clean_status(git_path=None):
    """
    Returns whether there are uncommitted changes in the working dir.
    """
    output = get_status(git_path)
    is_unmodified = (len(output.strip()) == 0)
    return is_unmodified

def get_last_commit_line(git_path=None):
    """
    Get one-line description of HEAD commit for repository in current dir.
    """
    output = call_git(git_path, "log", "--pretty=format:'%ad %h %s'",
                      "--date=short", "-n1")
    return output.strip()[1:-1]

def get_last_commit(git_path=None):
    """
    Get the HEAD commit SHA1 of repository in current dir.
    """
    line = get_last_commit_line(git_path)
    revision_id = line.split()[1]
    return revision_id

def print_summary(string='Repository', git_path=None):
    """
    Print the last commit line and eventual uncommitted changes.
    """
    # If git is available, check fretbursts version
    if not git_path_valid():
        print('\n%s revision unknown (git not found).' % string)
    else:
        last_commit = get_last_commit_line()
        print('\n{} revision:\n {}\n'.format(string, last_commit))
        if not check_clean_status():
            print('\nWARNING -> Uncommitted changes:')
            print(get_status())



#
# FRETBursts - A single-molecule FRET burst analysis toolkit.
#
# Copyright (C) 2014 Antonino Ingargiola <tritemio@gmail.com>
#
"""
Misc utility functions
"""

from __future__ import division, print_function
import os
import sys
import numpy as np


def _is_list_of_arrays(obj):
    return isinstance(obj, list) and np.all([isinstance(v, np.ndarray)
                                             for v in obj])

class HistData(object):
    """Stores histogram counts and bins and provides derived fields.

    Attributes:
        counts (array, ints): array of counts in each bin
        bins (array): array of bin edges. Size is size(counts) + 1.
        bincenters (array): array of bin  centers. Size is size(counts).
        pdf (array, floats): array of normalized counts (aka PDF)
    """
    def __init__(self, counts, bins):
        self.counts = counts
        self.bins = bins
        self.binwidth = bins[1] - bins[0]

    @property
    def bincenters(self):
        if not hasattr(self, '_bincenters'):
            self._bincenters = self.bins[:-1] + 0.5*self.binwidth
        return self._bincenters

    @property
    def pdf(self):
        if not hasattr(self, '_pdf'):
            self._pdf = np.array(self.counts, dtype=np.float)
            self._pdf /= (self.counts.sum() * self.binwidth)
        return self._pdf


def clk_to_s(t_ck, clk_p=12.5*1e-9):
    """Convert clock cycles to seconds."""
    return t_ck*clk_p

def pprint(s, mute=False):
    """Print immediately, even if inside a busy loop."""
    if mute: return
    sys.stdout.write(s)
    sys.stdout.flush()

def deprecate(function, old_name, new_name):
    def deprecated_function(*args, **kwargs):
        pprint("Function name %s is deprecated, use %s instead.\n" %\
                (old_name, new_name))
        res = function(*args, **kwargs)
        return res
    return deprecated_function

def shorten_fname(f):
    """Return a path with only the last subfolder (i.e. measurement date)."""
    return '/'.join(f.split('/')[-2:])

def binning(times, bin_width_ms=1, max_num_bins=1e5, clk_p=12.5e-9):
    """Return the binned histogram of array times."""
    bin_width_clk = (bin_width_ms*1e-3)/clk_p
    num_bins = min(times.max()/bin_width_clk, max_num_bins)
    h = np.histogram(times[times<(num_bins*bin_width_clk)], bins=num_bins)
    return h


def mkdir_p(path):
    """Create the path if not existent, otherwise do nothing.
    If `path` exists, and is not a dir, raise an exception.
    """
    import errno
    try:
        os.makedirs(path)
    except OSError as exc: # Python >2.5
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise exc

def download_file(url, save_dir='./'):
    """Download a file from `url` saving it to disk.

    The file name is taken from `url` and left unchanged.
    The destination dir can be set using `save_dir`
    (Default: the current dir).
    """
    ## Check if local path already exist
    fname = url.split('/')[-1]
    print('URL:  %s' % url)
    print('File: %s\n ' % fname)

    path = '/'.join([os.path.abspath(save_dir), fname])
    if os.path.exists(path):
        print('File already on disk: %s \nDelete it to re-download.' % path)
        return

    from future.standard_library import install_aliases
    install_aliases()
    from urllib.request import urlopen, urlretrieve
    from urllib.error import HTTPError, URLError

    ## Check if the URL is valid
    try:
        urlopen(url)
    except URLError as e:
        print('Wrong URL or no connection.\n\nError:\n%s\n' % e)
    except HTTPError:
        print('URL not found: ' + url)
        return

    ## Donwload the file
    def _report(blocknr, blocksize, size):
        current = blocknr*blocksize/2**20
        sys.stdout.write(
            "\rDownloaded {0:4.1f} / {1:4.1f} MB".format(current, size/2**20))
    mkdir_p(save_dir)
    urlretrieve(url, path, _report)




# Copyright (c) 2015-2017 Antonino Ingargiola
# License: MIT
"""
nbrun - Run an Jupyter/IPython notebook, optionally passing arguments.

USAGE
-----

Copy this file in the folder containing the master notebook used to
execute the other notebooks. Then use `run_notebook()` to execute
notebooks.
"""

import sys
import time
from pathlib import Path
from IPython.display import display, FileLink
import nbformat
from nbconvert.preprocessors import ExecutePreprocessor
from nbconvert import HTMLExporter

__version__ = '0.2'


def dict_to_code(mapping):
    """Convert input dict `mapping` to a string containing python code.

    Each key is the name of a variable and each value is
    the variable content. Each variable assignment is separated by
    a newline.

    Keys must be strings, and cannot start with a number (i.e. must be
    valid python identifiers). Values must be objects with a string
    representation (the result of repr(obj)) which is valid python code for
    re-creating the object.
    For examples, numbers, strings or list/tuple/dict
    of numbers and strings are allowed.

    Returns:
        A string containing the python code.
    """
    lines = ("{} = {}".format(key, repr(value))
             for key, value in mapping.items())
    return '\n'.join(lines)


def run_notebook(notebook_path, nb_kwargs=None, suffix='-out',
                 out_path_ipynb=None, out_path_html=None,
                 kernel_name=None, working_dir='./',
                 timeout=3600, execute_kwargs=None,
                 save_ipynb=True, save_html=False,
                 insert_pos=1, hide_input=False, display_links=True,
                 return_nb=False, add_header=True):
    """Runs a notebook and saves the output in a new notebook.

    Executes a notebook, optionally passing "arguments"
    similarly to passing arguments to a function.
    Notebook arguments are passed in a dictionary (`nb_kwargs`) which is
    converted into a string containing python assignments. This string is
    inserted in the template notebook as a code cell. The code assigns
    variables which can be used to control the execution. When "calling"
    a notebook, you need to know which arguments (variables) to pass.
    Unlike normal python functions, no check is performed on the input
    arguments. For sanity, we recommended describing the variables that
    can be assigned using a markdown cell at the beginning of the template
    notebook.

    Arguments:
        notebook_path (pathlib.Path or string): input notebook filename.
            This is the notebook to be executed (i.e. template notebook).
        nb_kwargs (dict or None): If not None, this dict is converted to a
            string of python assignments using the dict keys as variables
            names and the dict values as variables content. This string is
            inserted as code-cell in the notebook to be executed.
        suffix (string): suffix to append to the file name of the executed
            notebook. Argument ignored if `out_notebook_path` is not None.
        out_path_ipynb (pathlib.Path, string or None): file name for the
            output ipynb notebook. If None, the ouput ipynb notebook has
            the same name as the input notebook plus a suffix, specified
            by the `suffix` argument. If not None, `suffix` is ignored.
            If argument `save_ipynb` is False this argument is ignored.
        out_path_html (pathlib.Path, string or None): file name for the
            output HTML notebook. If None, the ouput HTML notebook has
            the same name as the input notebook plus a suffix, specified
            by the `suffix` argument. If not None, `suffix` is ignored.
            If argument `save_html` is False this argument is ignored.
        kernel_name (string or None): name of the kernel used to execute the
            notebook. Use the default kernel if None.
        working_dir (string or Path): the folder the kernel is started into.
        timeout (int): max execution time (seconds) for each cell before the
            execution is aborted.
        execute_kwargs (dict): additional arguments passed to
            `ExecutePreprocessor`.
        save_ipynb (bool): if True, save the output notebook in ipynb format.
            Default True.
        save_html (bool): if True, save the output notebook in HTML format.
            Default False.
        insert_pos (int): position of insertion of the code-cell containing
            the input arguments. Default is 1 (i.e. second cell). With this
            default, the input notebook can define, in the first cell, default
            values of input arguments (used when the notebook is executed
            with no arguments or through the Notebook GUI).
        hide_input (bool): whether to create a notebook with input cells
            hidden (useful to remind user that the auto-generated output
            is not meant to have the code edited.
        display_links (bool): if True, display/print "link" of template and
            output notebooks. Links are only rendered in a notebook.
            In a text terminal, links are displayed as full file names.
        return_nb (bool): if True, returns the notebook object. If False
            returns None. Default False.
    """
    timestamp_cell = ("---\n**Executed:** %s\n\n**Duration:** %d seconds.\n\n"
                      "**Autogenerated from:** [%s](%s)")
    if nb_kwargs is None:
        nb_kwargs = {}
    else:
        header = '# Cell inserted during automated execution.'
        code = dict_to_code(nb_kwargs)
        code_cell = '\n'.join((header, code))

    notebook_path = Path(notebook_path)
    if not notebook_path.is_file():
        raise FileNotFoundError("Path '%s' not found." % notebook_path)

    def check_out_path(notebook_path, out_path, ext, save):
        if out_path is None:
            out_path = Path(notebook_path.parent,
                            notebook_path.stem + suffix + ext)
        out_path = Path(out_path)
        if save and not out_path.parent.exists():
            msg = "Folder of the output %s file was not found:\n - %s\n."
            raise FileNotFoundError(msg % (ext, out_path_ipynb.parent))
        return out_path

    out_path_ipynb = check_out_path(notebook_path, out_path_ipynb,
                                    ext='.ipynb', save=save_ipynb)
    out_path_html = check_out_path(notebook_path, out_path_html,
                                   ext='.html', save=save_html)
    if display_links:
        display(FileLink(str(notebook_path)))

    if execute_kwargs is None:
        execute_kwargs = {}
    execute_kwargs.update(timeout=timeout)
    if kernel_name is None:
        kernel_name = 'python%d' % sys.version_info[0]
    execute_kwargs.update(kernel_name=kernel_name)
    ep = ExecutePreprocessor(**execute_kwargs)
    nb = nbformat.read(str(notebook_path), as_version=4)

    if hide_input:
        nb["metadata"].update({"hide_input": True})

    if len(nb_kwargs) > 0:
        nb['cells'].insert(insert_pos, nbformat.v4.new_code_cell(code_cell))

    start_time = time.time()
    try:
        # Execute the notebook
        ep.preprocess(nb, {'metadata': {'path': working_dir}})
    except:
        # Execution failed, print a message then raise.
        msg = ('Error executing the notebook "%s".\n'
               'Notebook arguments: %s\n\n'
               'See notebook "%s" for the traceback.' %
               (notebook_path, str(nb_kwargs), out_path_ipynb))
        print(msg)
        timestamp_cell += '\n\nError occurred during execution. See below.'
        raise
    finally:
        # Add timestamping cell
        duration = time.time() - start_time
        timestamp_cell = timestamp_cell % (time.ctime(start_time), duration,
                                           notebook_path, out_path_ipynb)
        if add_header:
            nb['cells'].append(nbformat.v4.new_markdown_cell(timestamp_cell))
        # Save the executed notebook to disk
        if save_ipynb:
            nbformat.write(nb, str(out_path_ipynb))
            if display_links:
                display(FileLink(str(out_path_ipynb)))
        if save_html:
            html_exporter = HTMLExporter()
            body, resources = html_exporter.from_notebook_node(nb)
            with open(str(out_path_html), 'w') as f:
                f.write(body)
        if return_nb:
            return nb


if __name__ == '__main__':
    import argparse

    descr = """\
        Execute all notebooks in a folder.
        """
    parser = argparse.ArgumentParser(description=descr, epilog='\n')
    parser.add_argument('folder',
                        help='Source folder with files to be processed.')
    msg = ('Name of kernel executing the notebook.\n'
           'Use `jupyter kernelspec list` for a list of kernels.')
    parser.add_argument('--kernel', metavar='KERNEL_NAME', default=None,
                        help=msg)
    parser.add_argument('--exclude-list', metavar='FILE_NAME', default=None,
                        help=msg)
    args = parser.parse_args()

    folder = Path(args.folder)
    assert folder.is_dir(), 'Folder "%s" not found.' % folder

    out_path = Path(folder, 'out/')
    if not out_path.is_dir():
        out_path.mkdir(parents=True)  # py2 compat

    exclude_list = []
    if args.exclude_list is not None and sys.version_info[0] == 2:
        with open(args.exclude_list) as f:
            exclude_list = [s.strip() for s in f.readlines()]

    print('Executing notebooks in "%s" ... ' % folder)
    pathlist = list(folder.glob('*.ipynb'))
    for nbpath in pathlist:
        if nbpath.name in exclude_list:
            print('- Skipping "%s"' % nbpath)
            continue
        if nbpath.stem.endswith('-out') or nbpath.stem.startswith('_'):
            continue
        print()
        out_path_ipynb = Path(out_path, nbpath.name)
        run_notebook(nbpath, out_path_ipynb=out_path_ipynb,
                     kernel_name=args.kernel)



#
# FRETBursts - A single-molecule FRET burst analysis toolkit.
#
# Copyright (C) 2014-2016 The Regents of the University of California,
#               Antonino Ingargiola <tritemio@gmail.com>
#

from __future__ import print_function, absolute_import

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions


## Citation information
_CITATION = """
   FRETBursts: An Open Source Toolkit for Analysis of Freely-Diffusing Single-Molecule FRET
   Ingargiola et al. (2016). http://dx.doi.org/10.1371/journal.pone.0160716 """

_INFO_CITATION = (' You are running FRETBursts (version {}).\n\n'
                  ' If you use this software please cite the following'
                  ' paper:\n{}\n\n').format(__version__, _CITATION)

def citation(bar=True):
    cit = _INFO_CITATION
    if bar:
        cit = ('-' * 62) + '\n' + _INFO_CITATION +  ('-' * 62)
    print(cit)


import warnings

try:
    import pandas
except ImportError:
    has_pandas = False
    warnings.warn((' - Cannot import pandas. Some functionality will not be '
                   'available.'))
else:
    has_pandas = True

try:
    import matplotlib
except ImportError:
    has_matplotlib = False
    warnings.warn((' - Cannot import matplotlib. Plotting will not be '
                   'available.'))
else:
    has_matplotlib = True

try:
    try:
        from PyQt5 import QtWidgets, QtCore
        QtGui = QtWidgets
    except ImportError:
        try:
            from PyQt4 import QtGui, QtCore
        except ImportError:
            from PySide import QtGui, QtCore
except ImportError:
    has_qt = False
    # This catches ImportError or other errors due to broken QT installation
    warnings.warn((' - Cannot import QT, custom GUI widgets disabled.'))
else:
    has_qt = True


try:
    import lmfit
except ImportError:
    has_lmfit = False
    warnings.warn((' - Cannot import lmfit. Some fitting functionalities '
                   ' will not be available.'))
else:
    has_lmfit = True


__all__numpy = ["np"]

__all__matplotlib = [
        # Library modules and functions
        "plt", "rcParams", "matplotlib", "plot", "hist",
        "grid", "xlim", "ylim", "gca", "gcf",]

__all_local_names = [
        # Local modules
        "loader", "select_bursts", "bl", "bg", "bpl", "bext", "bg_cache",
        "hdf5", "fretmath", "mfit", "citation", "git",

        # Classes, functions, variables
        "Data", "Sel", "Ph_sel",
        "download_file", "init_notebook",

        # Standalone plots or plots as a function of ch
        "mch_plot_bg", "plot_alternation_hist", "alex_jointplot",

        # Plots types used for 1ch of multi-ch plots through `dplot`
        "timetrace", "timetrace_single", "ratetrace", "ratetrace_single",
        "timetrace_fret", "timetrace_bg",
        "hist_width", "hist_size", "hist_size_all", "hist_brightness",
        "hist_fret", "hist_burst_data",
        "hist2d_alex", "hist_S", "hist_sbr", "hist_asymmetry",
        "hist_interphoton_single", "hist_interphoton",
        "hist_bg_single", "hist_bg", "hist_ph_delays", "hist_mdelays",
        "hist_mrates", "hist_burst_phrate", "hist_burst_delays",
        "scatter_width_size", "scatter_rate_da", "scatter_fret_size",
        "scatter_fret_nd_na", "scatter_fret_width", "scatter_da",
        "scatter_naa_nt", "scatter_alex", "hexbin_alex",

        # Wrapper functions that create a plot for each channel
        "dplot", "dplot_48ch", "dplot_8ch", "dplot_1ch",
        ]

__all__ = __all__numpy + __all_local_names

import numpy as np

if has_qt:
    __all__ += ['OpenFileDialog']
    from .utils.gui import OpenFileDialog

if has_matplotlib:
    __all__ += __all__matplotlib
    from matplotlib import rcParams
    import matplotlib.pyplot as plt
    from matplotlib.pyplot import plot, hist, grid, xlim, ylim, gca, gcf

# Import plain module names
from . import loader, hdf5, select_bursts, fretmath

# Import modules with custom names
from . import background as bg
from . import burstlib as bl

# Import objects
from .burstlib import Data, Sel
from .ph_sel import Ph_sel

if has_pandas and has_lmfit:
    from . import burstlib_ext as bext

if has_matplotlib and has_pandas and has_lmfit:
    from . import mfit
    from . import burst_plot as bpl
    from .burst_plot import (
            # Standalone plots as a function of ch
            mch_plot_bg, plot_alternation_hist, alex_jointplot,

            # Single-ch plots used in multi-ch plots through `dplot`
            timetrace, timetrace_single, ratetrace, ratetrace_single,
            timetrace_fret, timetrace_bg,
            hist_width, hist_size, hist_size_all, hist_brightness,
            hist_fret, hist_burst_data,
            hist2d_alex, hist_S, hist_sbr, hist_asymmetry,
            hist_interphoton_single, hist_interphoton,
            hist_bg_single, hist_bg, hist_ph_delays, hist_mdelays,
            hist_mrates, hist_burst_phrate, hist_burst_delays,
            scatter_width_size, scatter_rate_da, scatter_fret_size,
            scatter_fret_nd_na, scatter_fret_width, scatter_da,
            scatter_naa_nt, scatter_alex, hexbin_alex,

            # Wrapper functions that create a plot for each channel
            dplot, dplot_48ch, dplot_8ch, dplot_1ch,
            )

from .utils.misc import download_file
from .utils import git


def init_notebook(fs=13, seaborn_style='darkgrid',
                  mpl_backend='inline', configure_logger=True, apionly=False):
    """
    Set the default plot style for inline plots using the seaborn library.

    This function must be called from an ipython notebook or
    ipython QT console.

    Arguments:
        fs (int): base font size for text labels

    Returns:
        The imported seaborn library. By saving the return value you
        don't need to import seaborn again.
    """
    if configure_logger:
        import logging
        logger = logging.getLogger('fretbursts')
        formatter = logging.Formatter('%(levelname)s %(name)s\n%(message)s')
        handler = logging.StreamHandler()
        handler.setFormatter(formatter)
        logger.addHandler(handler)

    if mpl_backend is not None:
        ip = get_ipython()
        ip.enable_matplotlib(mpl_backend)

    import seaborn as sns
    if not apionly:
        rc={'font.size': fs, 'axes.labelsize': fs, 'legend.fontsize': fs,
            'axes.titlesize': fs*1.1,
            'xtick.labelsize': fs, 'ytick.labelsize': fs,
            'font.sans-serif': ['Arial', 'Liberation Sans'],
        }
        sns.set(rc=rc)
        blue = '#0055d4'
        green = '#2ca02c'
        color_brewer = sns.color_palette("Set1", 9)
        colors = np.array(color_brewer)[(1,0,2,3,4,8,6,7), :]
        colors = list(colors)
        colors[:3] = (blue, colors[1], green)
        sns.set_palette(colors, 8)
        sns.colors = colors
        sns.set_style(seaborn_style)
    return sns

citation()



from gaussian_fitting import *

mu0 = -0.1
sigma0 = 1.8
print("Real values: %s %s" % (mu0, sigma0))

## Compare the two methods
n = 10000
Mu, Sigma = zeros(n), zeros(n)
Muh, Sigmah = zeros(n), zeros(n)
Mu1, Sigma1 = zeros(n), zeros(n)
for i in range(n):
    s = R.normal(size=100, loc=mu0, scale=sigma0)
    #print "Empirical mean and std", s.mean(), s.std()
    mu, sigma =  gaussian_fit(s, [0,1])
    muh, sigmah, h, bins =  gaussian_fit_hist(s, [0,1])
    mu1, sigma1 = s.mean(), s.std()

    Mu[i] = mu; Sigma[i] = sigma
    Muh[i] = muh; Sigmah[i] = sigmah
    Mu1[i] = mu1; Sigma1[i] = sigma1

figure(1); title("Mu")
b = r_[mu0-1:mu0+1:40j]
hist(Mu, bins=b, alpha=0.3, label='ECDF')
hist(Muh, bins=b, alpha=0.3, color='r', label='Histo')
hist(Mu1, bins=b, alpha=0.3, color='y', label='Moments')
legend(loc='best')
axvline(mu0, color='k', lw=2)
grid(True)

figure(2); title("Sigma")
b = r_[sigma0-1:sigma0+1:40j]
hist(Sigma, bins=b, alpha=0.3, label='ECDF')
hist(Sigmah, bins=b, alpha=0.3, color='r', label='Histo')
hist(Sigma1, bins=b, alpha=0.3, color='y', label='Moments')
legend(loc='best')
axvline(sigma0, color='k', lw=2)
grid(True)

show()





#
# FRETBursts - A single-molecule FRET burst analysis toolkit.
#
# Copyright (C) 2017 Antonino Ingargiola <tritemio@gmail.com>
#
"""
Unit tests for Burst, Bursts, BurstGap, BurstsGap.

Running the tests requires `py.test`.
"""

from __future__ import division
from builtins import range, zip

import numpy as np
import pytest

from fretbursts.phtools import burstsearch as bs


def test_Burst():
    burst = bs.Burst(0, 1, 100, 200)
    assert burst.counts == 2
    assert burst.width == 100
    assert burst.ph_rate == 2 / 100


def test_BurstGap():
    burst = bs.BurstGap(0, 2, 100, 300, 100, 1)
    assert burst.counts == 2
    assert burst.width == 100
    assert burst.gap == 100
    assert burst.gap_counts == 1


def test_Bursts():
    barray = np.zeros((100, 4), dtype=np.int64)
    barray[:, 0] = np.arange(100) * 10
    barray[:, 1] = barray[:, 0] + 5
    barray[:, 2] = np.arange(100) * 100
    barray[:, 3] = barray[:, 2] + 50

    bursts = bs.Bursts(barray)
    assert bursts.num_bursts == 100
    assert (bursts.counts == 6).all()
    assert (bursts.width == 50).all()

    # Test indexing and slicing
    assert bursts[0] == bs.Bursts(barray[0])
    assert bursts[:1] == bursts[0]
    assert (bursts[:10:2].data == barray[:10:2]).all()


def test_BurstsGap():
    barray = np.zeros((100, 6), dtype=np.int64)
    barray[:, 0] = np.arange(100) * 10
    barray[:, 1] = barray[:, 0] + 6
    barray[:, 2] = np.arange(100) * 100
    barray[:, 3] = barray[:, 2] + 80
    barray[:, 4] = 30
    barray[:, 5] = 1

    bursts = bs.BurstsGap(barray)
    assert bursts.num_bursts == 100
    assert (bursts.counts == 6).all()
    assert (bursts.width == 50).all()

    # Test indexing and slicing
    assert bursts[0] == bs.BurstsGap(barray[0])
    assert bursts[:1] == bursts[0]
    assert (bursts[:10:2].data == barray[:10:2]).all()


if __name__ == '__main__':
    pytest.main("-x -v fretbursts/tests/test_Bursts.py")



#
# FRETBursts - A single-molecule FRET burst analysis toolkit.
#
# Copyright (C) 2014 Antonino Ingargiola <tritemio@gmail.com>
#
"""
The `loader` module contains functions to load each supported data format.
The loader functions load data from a specific format and
return a new :class:`fretbursts.burstlib.Data()` object containing the data.

This module contains the high-level function to load a data-file and
to return a `Data()` object. The low-level functions that perform the binary
loading and preprocessing can be found in the `dataload` folder.
"""

from __future__ import print_function, absolute_import
from builtins import range, zip

import os
import pickle
import numpy as np
import tables

from .dataload.multi_ch_reader import load_data_ordered16
from .dataload.manta_reader import (load_manta_timestamps,
                                    load_xavier_manta_data,
                                    get_timestamps_detectors,
                                    # process_timestamps,
                                    process_store,
                                    load_manta_timestamps_pytables)
from .utils.misc import pprint, deprecate
from .burstlib import Data
from .hdf5 import hdf5_data_map

from phconvert.hdf5 import dict_from_group


def multispot8(fname, bytes_to_read=-1, swap_D_A=True, leakage=0, gamma=1.):
    """Load a 8-ch multispot file and return a Data() object. Cached version.
    """
    fname_c = fname + '_cache.pickle'
    try:
        var = pickle.load(open(fname_c, 'rb'))
        dx = Data(fname=fname, clk_p=12.5e-9, nch=8, leakage=leakage,
                  gamma=gamma)
        dx.add(ph_times_m=var['ph_times_m'], A_em=var['A_em'], ALEX=False)
        pprint(" - File loaded from cache: %s\n" % fname)
    except IOError:
        dx = multispot8_core(fname, bytes_to_read=bytes_to_read,
                             swap_D_A=swap_D_A, leakage=leakage, gamma=gamma)
        D = {'ph_times_m': dx.ph_times_m, 'A_em': dx.A_em}
        pprint(" - Pickling data ... ")
        pickle.dump(D, open(fname_c, 'wb'), -1)
        pprint("DONE\n")
    return dx

load_multispot8 = deprecate(multispot8, "load_multispot8", "loader.multispot8")


def multispot8_core(fname, bytes_to_read=-1, swap_D_A=True, leakage=0,
                    gamma=1.):
    """Load a 8-ch multispot file and return a Data() object.
    """
    dx = Data(fname=fname, clk_p=12.5e-9, nch=8, leakage=leakage,
              gamma=gamma)
    ph_times_m, A_em, ph_times_det = load_data_ordered16(
        fname=fname, n_bytes_to_read=bytes_to_read, swap_D_A=swap_D_A)
    dx.add(ph_times_m=ph_times_m, A_em=A_em, ALEX=False)
    return dx


def multispot48_simple(fname, leakage=0, gamma=1.,
                       i_start=0, i_stop=None, debug=False):
    """Load a 48-ch multispot file and return a Data() object.
    """
    dx = Data(fname=fname, clk_p=10e-9, nch=48, leakage=leakage, gamma=gamma)
    ph_times_m, big_fifo, ch_fifo = load_manta_timestamps(
        fname, i_start=i_start, i_stop=i_stop, debug=debug)
    A_em = [True] * len(ph_times_m)
    dx.add(ph_times_m=ph_times_m, A_em=A_em, ALEX=False)
    big_fifo_full = np.array([b.any() for b in big_fifo]).any()
    ch_fifo_full = np.array([b.any() for b in ch_fifo]).any()
    if big_fifo_full:
        print('WARNING: Big-FIFO full, flags saved in Data()')
        dx.add(big_fifo=big_fifo)
    if ch_fifo_full:
        print('WARNING: CH-FIFO full, flags saved in Data()')
        dx.add(ch_fifo=ch_fifo)
    return dx


def multispot48(fname, leakage=0, gamma=1., reprocess=False,
                i_start=0, i_stop=None, debug=False):
    """Load a 48-ch multispot file and return a Data() object.
    """
    basename, ext = os.path.splitext(fname)
    fname_h5 = basename + '.hdf5'
    fname_dat = basename + '.dat'

    def load_dat_file():
        pprint(' - Loading DAT file: %s ... ' % fname_dat)
        # Load data from raw file and store it in a HDF5 file
        data = load_xavier_manta_data(fname_dat, i_start=i_start,
                                      i_stop=i_stop, debug=debug)
        pprint('DONE.\n - Extracting timestamps and detectors ... ')
        timestamps, det = get_timestamps_detectors(data, nbits=24)
        pprint('DONE.\n - Processing and storing ... ')
        ph_times_m, big_fifo, ch_fifo = process_store(
            timestamps, det, out_fname=fname_h5, fifo_flag=True, debug=False)
        pprint('DONE.\n')
        return ph_times_m, big_fifo, ch_fifo

    if not (os.path.isfile(fname_dat) or os.path.isfile(fname_h5)):
        raise IOError('Data file "%s" not found' % basename)

    if os.path.exists(fname_h5) and not reprocess:
        # There is a HDF5 file
        try:
            pprint(' - Loading HDF5 file: %s ... ' % fname_h5)
            ph_times_m, big_fifo, ch_fifo = \
                load_manta_timestamps_pytables(fname_h5)
            pprint('DONE.\n')
        except tables.HDF5ExtError:
            pprint('\n  Ops! File may be truncated.\n')
            ph_times_m, big_fifo, ch_fifo = load_dat_file()
    else:
        ph_times_m, big_fifo, ch_fifo = load_dat_file()

    # Current data has only acceptor ch
    A_em = [True] * len(ph_times_m)

    dx = Data(fname=fname, clk_p=10e-9, nch=48, leakage=leakage, gamma=gamma)
    dx.add(ph_times_m=ph_times_m, A_em=A_em, ALEX=False,
           data_file=ph_times_m.data_file, bg_data_file=ph_times_m.data_file)
    big_fifo_full = np.array([b[:].any() for b in big_fifo]).any()
    ch_fifo_full = np.array([b[:].any() for b in ch_fifo]).any()
    if big_fifo_full:
        print('WARNING: Big-FIFO full, flags saved in Data()')
        dx.add(big_fifo=big_fifo)
    if ch_fifo_full:
        print('WARNING: CH-FIFO full, flags saved in Data()')
        dx.add(ch_fifo=ch_fifo)
    return dx


def assert_valid_photon_hdf5(h5file):
    meta = dict(format_name=b'Photon-HDF5', format_version=b'0.2')
    msg = ''
    for attr, value in meta.items():
        if attr not in h5file.root._v_attrs:
            msg += ' * Error: %s not in %s\n' % (attr, h5file.root._v_attrs)
        stored_value = h5file.root._v_attrs[attr]
        if stored_value != value:
            msg += ' * Error: %s != %s\n' % (stored_value, value)
    if msg is not '':
        h5file.close()
        msg = 'Not a valid Photon-HDF5 file. \n' + msg
        raise IOError(msg)


def _is_basic_layout(h5file):
    return 'photon_data' in h5file.root


class H5Loader():
    def __init__(self, h5file, data):
        self.h5file = h5file
        self.data = data

    def load_data(self, where, name, dest_name=None, ich=None,
                  allow_missing=False, ondisk=False):
        try:
            node = self.h5file.get_node(where, name)
        except tables.NoSuchNodeError:
            if allow_missing:
                node_value = np.array([])
            else:
                self.h5file.close()
                raise IOError("Invalid file format: '%s' is missing." % name)
        else:
            node_value = node if ondisk else node.read()

        if dest_name is None:
            dest_name = hdf5_data_map.get(name, name)

        if ich is None:
            self.data.add(**{dest_name: node_value})
        else:
            if dest_name not in self.data:
                self.data.add(**{dest_name: [node_value]})
            else:
                self.data[dest_name].append(node_value)


def hdf5(fname, ondisk=False):
    """Load a data file saved in Photon-HDF5 format, version 0.2 only.

    The format version 0.2 is obsolete, please use :func:`photon_hdf5`
    for Photon-HDF5 version 0.3 or higher. If you have files in
    Photon-HDF5 format 0.2 please convert them to version 0.3 or higher.

    For description and specs of the Photon-HDF5 format see:
    http://photon-hdf5.readthedocs.org/

    Arguments:
        ondisk (bool): if True do not load in the timestamp arrays, just
            load the array reference. Multi-spot only. Default False.

    Returns:
        :class:`fretbursts.burstlib.Data` object containing the data.
    """
    print('WARNING: You are using Photon-HDF5 version 0.2 which is '
          'obsolete. Please update you file to version 0.3 or higher.')

    # This used for reading Photon-HDF5 version 0.2.
    mandatory_root_fields = ['timestamps_unit', 'num_spots', 'num_detectors',
                             'num_spectral_ch', 'num_polariz_ch',
                             'alex', 'lifetime']

    if not os.path.isfile(fname):
        raise IOError('File not found.')
    data_file = tables.open_file(fname, mode="r")
    assert_valid_photon_hdf5(data_file)

    d = Data(fname=fname)
    loader = H5Loader(data_file, d)

    # Load mandatory parameters
    mand_fields = [f for f in mandatory_root_fields if f != 'num_detectors']
    for field in mand_fields:
        loader.load_data('/', field)
    try:
        data_file.get_node('/', 'num_detectors')
    except tables.NoSuchNodeError:
        pprint("WARNING: No 'num_detectors' field found in the HDF5 file.\n")
    else:
        loader.load_data('/', 'num_detectors')

    if d.ALEX:
        if d.lifetime:
            loader.load_data('/', 'laser_pulse_rate', allow_missing=True)
        else:
            loader.load_data('/', 'alex_period')
        loader.load_data('/', 'alex_period_donor', allow_missing=True)
        loader.load_data('/', 'alex_period_acceptor', allow_missing=True)

    ## Load metadata groups
    metagroups = ['/setup', '/provenance', '/identity']
    for group in metagroups:
        if group in data_file:
            d.add(**{group[1:]: dict_from_group(data_file.get_node(group))})

    if '/acquisition_time' in data_file:
        d.add(acquisition_time=data_file.root.acquisition_time.read())

    if _is_basic_layout(data_file):
        ph_group = data_file.root.photon_data

    if d.lifetime:
        try:
            assert 'nanotimes' in ph_group
            assert 'nanotimes_specs' in ph_group
        except AssertionError:
            data_file.close()
            raise IOError(('The lifetime flag is True but the TCSPC '
                           'data is missing.'))

    if d.nch == 1:
        # load single-spot data from "basic layout"
        if not d.ALEX:
            mapping = {'timestamps': 'ph_times_m', 'detectors': 'A_em',
                       'nanotimes': 'nanotimes', 'particles': 'particles'}
            ich = 0  # Created a 1-element list for each field
        else:
            mapping = {'timestamps': 'ph_times_t', 'detectors': 'det_t',
                       'nanotimes': 'nanotimes_t', 'particles': 'particles_t'}
            ich = None  # don't warp the arrays in a list
        for name, dest_name in mapping.items():
            if name in ph_group:
                loader.load_data(ph_group, name, dest_name=dest_name, ich=ich)

        if 'detectors_specs' in ph_group:
            det_specs = ph_group.detectors_specs
            if 'donor' in det_specs and 'acceptor' in det_specs:
                donor = det_specs.donor.read()
                accept = det_specs.acceptor.read()
                d.add(det_donor_accept=(donor, accept))
                if not d.ALEX:
                    # We still have to convert A_em to a boolean mask
                    # because until now it is just the detector list

                    # Make sure there are at most 2 detectors
                    assert len(np.unique(d.A_em[0])) <= 2
                    if accept and not donor:
                        # In this case we can convert without a copy
                        d.add(A_em=[d.A_em[0].view(dtype=bool)])
                    else:
                        # Create the boolean mask
                        d.add(A_em=[d.A_em[0] == accept])

        if 'nanotimes_specs' in ph_group:
            nanot_specs = ph_group.nanotimes_specs
            nanotimes_params = {}
            for name in ['tcspc_unit', 'tcspc_num_bins', 'tcspc_range']:
                value = nanot_specs._f_get_child(name).read()
                nanotimes_params.update(**{name: value})
            for name in ['tau_accept_only', 'tau_donor_only',
                         'tau_fret_donor', 'inverse_fret_rate']:
                if name in nanot_specs:
                    value = nanot_specs._f_get_child(name).read()
                    nanotimes_params.update(**{name: value})
            d.add(nanotimes_params=nanotimes_params)

    else:
        # Load multi-spot data from multi-spot layout
        for ich in range(d.nch):
            ph_group_name = '/photon_data_%d' % ich
            loader.load_data(ph_group_name, 'timestamps', allow_missing=True,
                             dest_name='ph_times_m', ich=ich, ondisk=ondisk)

            if ph_group_name not in data_file:
                a_em = np.array([], dtype=bool)
            elif ph_group_name + '/detectors' not in data_file:
                a_em = slice(None)
            else:
                ph_group = data_file.root._f_get_child(ph_group_name)
                det_specs = ph_group.detectors_specs
                donor = det_specs.donor.read()
                accept = det_specs.acceptor.read()
                if ph_group.detectors.dtype == np.bool:
                    a_em = ph_group.detectors.read()
                    if not accept:
                        np.logical_not(a_em, out=a_em)
                else:
                    det = ph_group.detectors.read()
                    a_em = (det == accept)
                    d_em = (det == donor)
                    assert not (a_em * d_em).any()
                    assert (a_em + d_em).all()

            if 'A_em' not in d:
                d.add(A_em=[a_em])
            else:
                d.A_em.append(a_em)

    d.add(data_file=data_file)
    return d
